/* libiriverdb - database manipulation routines
 * Copyright 2005 Matt Dainty
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * $Id: db.c 32 2005-08-14 19:28:01Z  $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/param.h>
#include <assert.h>
#include <errno.h>
#include <glib.h>
#include <glib/gi18n-lib.h>

#include "iriverdb.h"

#include "internal.h"

#define DB_FILENAME	"iRivNavi.iDB"
#define TRACK_MAX	10240
#define BOM		0xfeff

#ifdef DEBUG
static char *field[] = {
	"filename",
	"songname",
	"artist",
	"album",
	"genre",
};
#endif

/* Jump table of routines for doing per-fileformat metadata lookups
 */
extern iriverdb_format_t formats[];

/* Comparison function used by g_list_sort() for sorting the tracks
 *
 * FIXME Broken for UTF-8 strings
 */
gint iriverdb_sort( gconstpointer a, gconstpointer b )
{
#ifdef DEBUG
	/*fprintf( stderr, "%s: \"%s\" <=> \"%s\"\n",
			 __func__,
			 ( ( iriverdb_data_t * ) a )->field[IRIVERDB_FIELD_FILENAME],
			 ( ( iriverdb_data_t * ) b )->field[IRIVERDB_FIELD_FILENAME] );*/
#endif /* DEBUG */
	return g_utf8_collate( ( ( iriverdb_data_t * ) a )->field[IRIVERDB_FIELD_FILENAME],
			       ( ( iriverdb_data_t * ) b )->field[IRIVERDB_FIELD_FILENAME] );
}

/* Comparison function used by g_list_find_custom() to check for duplicate
 * entries
 *
 * FIXME Should sanitise filenames before comparison, otherwise '\foo\bar'
 *	 and '\foo\\bar' are not considered equivalent. Probably do
 *	 case-insensitive comparison too.
 */
gint iriverdb_find( gconstpointer a, gconstpointer b )
{
#ifdef DEBUG
	/*fprintf( stderr, "%s: \"%s\" == \"%s\"\n",
			 __func__, ( char * ) b,
			 ( ( iriverdb_data_t * ) a )->field[IRIVERDB_FIELD_FILENAME] );*/
#endif /* DEBUG */
	return strcasecmp( ( char * ) b,
			   ( ( iriverdb_data_t * ) a )->field[IRIVERDB_FIELD_FILENAME] );
}

/* Recurse over a given filesystem root adding any file not already in the
 * given list
 */
static int iriverdb_recurse( GList **list,
			     const char *root,
			     const char *path,
			     void ( *callback )( const iriverdb_callback_t type,
						 const void *data ) )
{
	GDir *dir;
	const gchar *entry;
	char buffer[PATH_MAX], *ext, *filename, *pathsep, *temp;
	struct stat buf;
	iriverdb_data_t *data;
	iriverdb_format_t *ptr;
	int i, count = 0;
	const char *unknown = N_( "Unknown" );

	if ( ( dir = g_dir_open( path, 0, NULL ) ) == NULL )
		return 0;

	while( ( entry = g_dir_read_name( dir ) ) != NULL ) {

		/* Ignore hidden, parent and current directory entries
		 */
		if ( strncmp( entry, ".", 1 ) == 0 )
			continue;

		/* FIXME Should probably sanitise the filename here, remove
		 *	 any double slashes, etc.
		 */
		sprintf( buffer, "%s%c%s",
				 path,
				 G_DIR_SEPARATOR,
				 entry );

		if ( stat( buffer, &buf ) <  0 ) {
			g_dir_close( dir );
			return count;
		}

		/* Recurse through child directories
		 */
		if ( S_ISDIR( buf.st_mode ) )
			count += iriverdb_recurse( list, root, buffer, callback );

		if ( ! S_ISREG( buf.st_mode ) )
			continue;

		/* Skip over files with no extension
		 *
		 * XXX perhaps use MIME types?
		 */
		if ( ( ext = rindex( entry, '.' ) ) == NULL )
			continue;
		
		/* This should point to the path, minus the root
		 */
		filename = buffer + strlen( root );
		temp = ( char * ) calloc( strlen( filename ) + 1, sizeof( char ) );
		strcpy( temp, filename );

#ifndef G_OS_WIN32
		/* Now turn all the path separators from '/' into '\'
		 */
		while ( ( pathsep = strchr( temp, G_DIR_SEPARATOR ) ) != NULL ) {
			*pathsep = '\\';
		}
#endif /* ! G_OS_WIN32 */

		/* Check that this filename doesn't already exist in the
		 * database
		 */
		if ( g_list_find_custom( *list, ( gconstpointer ) temp, iriverdb_find ) != NULL ) {
			free( temp );
			continue;
		}

		/* TODO */

		for( ptr = formats; ptr->ext != NULL; ptr++ ) {
			if( strcasecmp( ext, ptr->ext ) != 0 )
				continue;

			if ( ( ptr->info == NULL ) ||
			     ( ( data = ptr->info( buffer ) ) == NULL ) )
				continue;

			/* Call the callback
			 */
			if ( callback )
				callback( IRIVERDB_CALLBACK_SCAN, ( const void * ) buffer );

			/* Replace any NULL or empty field with
			 * (i18n) "Unknown"
			 */
			for( i = 1; i < IRIVERDB_FIELD_MAX; i++ ) {
				if ( ( data->field[i] == NULL ) ||
				     ( strlen( data->field[i] ) == 0 ) ) {
					if ( data->field[i] != NULL )
						free( data->field[i] );
					
					data->field[i] = ( char * ) calloc( strlen( iriverdb_gettext( unknown ) ) + 1, sizeof( char ) );
					strcpy( data->field[i], iriverdb_gettext( unknown ) );
				}
			}

			data->field[IRIVERDB_FIELD_FILENAME] = temp;

			*list = g_list_append( *list, data );
			count++;

			break;
		}

		/* The file didn't already exist in the database, we tried
		 * all of the format comparisons, none matched, so this wasn't
		 * used
		 */
		if ( ptr->ext == NULL )
			free( temp );
	}

	g_dir_close( dir );

	return count;
}

/* Write out a value as 16-bit little-endian to a stream
 */
static void iriverdb_write_le16( FILE *file, int value )
{
	int i;

	for( i = 0; i < 2; i++, value >>= 8 )
		fputc( value & 0xff, file );
}

/* Write out a value as 32-bit little-endian to a stream
 */
static void iriverdb_write_le32( FILE *file, int value )
{
	int i;

	for( i = 0; i < 4; i++, value >>= 8 )
		fputc( value & 0xff, file );
}

/* Read a 16-bit little endian value from a stream
 */
static int iriverdb_read_le16( FILE *file )
{
	int i, shift, value;

	for( i = shift = value = 0; i < 2; i++, shift += 8 )
		value |= ( fgetc( file ) << shift );

	return value;
}

/* Read a 32-bit little endian value from a stream
 */
static int iriverdb_read_le32( FILE *file )
{
	int i, shift, value;

	for( i = shift = value = 0; i < 4; i++, shift += 8 )
		value |= ( fgetc( file ) << shift );

	return value;
}

/* Return the total on-disk length of a string given the intended version of
 * the database
 */
static int iriverdb_strlen( char *string, iriverdb_version_t version )
{
	unsigned int size;
	gchar *temp;

	switch( version ) {
		case IRIVERDB_012:
			size = strlen( string ) + 1;
			break;
		case IRIVERDB_016:
			temp = iriverdb_utf8_to_utf16le( string, &size );
			size += 4;
			g_free( temp );
			break;
	}

	return size;
}

/* Write a string to a stream in the format required by the given database
 * version
 */
static void iriverdb_write_string( FILE *db,
				   char *string,
				   iriverdb_version_t version )
{
	gchar *temp;
	gsize size;

	switch( version ) {
		case IRIVERDB_012:
			fwrite( string, sizeof( char ),
				iriverdb_strlen( string, version ), db );
			break;
		case IRIVERDB_016:
			temp = iriverdb_utf8_to_utf16le( string, &size );
			iriverdb_write_le16( db, BOM );
			fwrite( temp, sizeof( char ), size, db );
			iriverdb_write_le16( db, 0 );
			g_free( temp );
			break;
	}
}

/* Given the database version and total on-disk length, read a string from
 * a stream
 */
static char *iriverdb_read_string( FILE *db,
				   int length,
				   iriverdb_version_t version )
{
	int i;
	char *temp, *result;

	/* Skip BOM
	 */
	if ( version == IRIVERDB_016 ) {
		fseek( db, 2, SEEK_CUR );
		length -= 2;
	}

	temp = ( char * ) g_malloc0( length * sizeof( char ) );
	fread( temp, sizeof( char ), length, db );

	/* If this is a 0.12 database, just return the string, it's already
	 * either ASCII or UTF-8
	 */
	if ( version == IRIVERDB_012 ) {
		return temp;
	}

	result = iriverdb_utf16le_to_utf8( temp, NULL );
	g_free( temp );

	return result;
}

/* Do the actual writing of a database to disk
 */
static int iriverdb_write_database( const char *path,
				    iriverdb_t *database,
				    void ( *callback )( const iriverdb_callback_t type,
							const void *data ) )
{
	char filename[PATH_MAX];
	FILE *db;
	int i, j, size, offset;
	iriverdb_data_t *data;

	sprintf( filename, "%s%c%s", path, G_DIR_SEPARATOR, DB_FILENAME );

	if ( ( db = fopen( filename, "wb" ) ) == NULL ) {
		return -1;
	}

	/* Zero out the fixed-size part of the database
	 */
	for( i = 0; i < 0xA0C0; i++ ) {
		fputc( 0, db );
	}

	/* Create the header
	 */
	fseek( db, 0, SEEK_SET );
	switch( database->version ) {
		case IRIVERDB_012:
			fputs( "iRivDB Ver 0.12", db );
			break;
		case IRIVERDB_016:
			fputs( "iRivDB Ver 0.16", db );
			break;
	}
	fseek( db, 0x0020, SEEK_SET );
	fputs( "iRiver iHP-100 DB File", db );

	/* Number of files in the database
	 */
	fseek( db, 0x0040, SEEK_SET );
	iriverdb_write_le32( db, g_list_length( database->tracks ) );
#ifdef DEBUG
	fprintf( stderr, "%s: total tracks = %d\n",
			 __func__, g_list_length( database->tracks ) );
#endif /* DEBUG */

	/* Write the "Designed by..." header
	 */
	fseek( db, 0xA080, SEEK_SET );
	fputs( "Designed by iRiver", db );

	/* Sort the database
	 */
	database->tracks = g_list_sort( database->tracks, iriverdb_sort );

	/* XXX Looking at the database format, there appears to be a limit of
	 *     10240 files in the TOC/FAT, so use that as an upper bound
	 *
	 * Write out the TOC and data
	 */
	for( i = offset = 0; i < g_list_length( database->tracks ) && i < TRACK_MAX; i++ ) {

		data = g_list_nth_data( database->tracks, i );

#ifdef DEBUG
		fprintf( stderr, "%s: track %d = \"%s\"\n",
				 __func__, i + 1,
				 data->field[IRIVERDB_FIELD_FILENAME] );
#endif /* DEBUG */

		/* Write the TOC entry
		 */
		fseek( db, 0x0080 + ( i << 2 ), SEEK_SET );
		iriverdb_write_le32( db, offset );

		fseek( db, 0, SEEK_END );
		offset += 10;

		/* Write the length of the fields
		 */
		for( j = 0; j < IRIVERDB_FIELD_MAX; j++ ) {
			size = iriverdb_strlen( data->field[j],
						database->version );
			iriverdb_write_le16( db, size );
			offset += size;
		}

		/* Now write the actual fields
		 */
		for( j = 0; j < IRIVERDB_FIELD_MAX; j++ )
			iriverdb_write_string( db, data->field[j],
					       database->version );
	}

	/* Pad out the database length to a multiple of 64 bytes
	 */
	fseek( db, 0, SEEK_END );
	for( i = 0; i < ( ftell( db ) & 0x3f ); i++ ) {
		fputc( 0, db );
	}
	
	fclose( db );

	return g_list_length( database->tracks );
}

/* Free all memory allocated within a database structure
 */
void iriverdb_free_database( iriverdb_t *database )
{
	int i, j;
	iriverdb_data_t *data;

	for( i = 0; i < g_list_length( database->tracks ); i++ ) {
		data = g_list_nth_data( database->tracks, i );
		for( j = 0; j < IRIVERDB_FIELD_MAX; j++ )
			free( data->field[j] );
		free( data );
	}

	g_list_free( database->tracks );

	free( database );
}

/* Create a new database in the given filesystem root for the specified
 * version with no callback for progress feedback
 */
int iriverdb_write( const char *path, iriverdb_version_t version )
{
	return iriverdb_write_with_callback( path, version, NULL );
}

/* Create a new database in the given filesystem root for the specified
 * version using the given callback to supply progress feedback
 */
int iriverdb_write_with_callback( const char *path,
				  iriverdb_version_t version,
				  void ( *callback )( const iriverdb_callback_t type,
						      const void *data ) )
{
	int total;
	iriverdb_t *database;

	if ( ( database = ( iriverdb_t * ) calloc( sizeof( iriverdb_t ), 1 ) ) == NULL )
		return -1;

	database->version = version;

	total = iriverdb_recurse( &database->tracks, path, path, callback );
	assert( total == g_list_length( database->tracks ) );

	if ( total > 0 )
		iriverdb_write_database( path, database, callback );

	iriverdb_free_database( database );

	return total;
}

/* Read in the database found in the given filesystem root with no callback
 * for progress feedback
 */
iriverdb_t *iriverdb_read( const char *path )
{
	return iriverdb_read_with_callback( path, NULL );
}

/* Read in the database found in the given filesystem root using the given
 * callback to supply progress feedback
 */
iriverdb_t *iriverdb_read_with_callback( const char *path,
					 void ( *callback )( const iriverdb_callback_t type,
							     const void *data ) )
{
	char filename[PATH_MAX], version[32], *ptr;
	FILE *db;
	iriverdb_t *database;
	int i, j, k, total, offset, field_length[IRIVERDB_FIELD_MAX];
	iriverdb_data_t *data;
	GList *toc = NULL;

	sprintf( filename, "%s%c%s", path, G_DIR_SEPARATOR, DB_FILENAME );

	if ( ( db = fopen( filename, "rb" ) ) == NULL ) {
		return NULL;
	}

	database = ( iriverdb_t * ) calloc( sizeof( iriverdb_t ), 1 );

	/* Work out what version of the database we're dealing with
	 */
	if ( fread( version, sizeof( char ), 32, db ) != 32 ) {
		free( database );
		return NULL;
	}

	if ( strcmp( version, "iRivDB Ver 0.12" ) == 0 ) {
		database->version = IRIVERDB_012;
	} else if ( strcmp( version , "iRivDB Ver 0.16" ) == 0 ) {
		database->version = IRIVERDB_016;
	} else {
		free( database );
		return NULL;
	}

	/* Read in the total number of tracks
	 */
	fseek( db, 0x0040, SEEK_SET );
	total = iriverdb_read_le32( db );

#ifdef DEBUG
	fprintf( stderr, "%s: total tracks = %d\n", 
			 __func__, total );
#endif /* DEBUG */

	/* Read in TOC
	 */
	fseek( db, 0x0080, SEEK_SET );
	for( i = 0; i < total; i++ ) {
		offset = iriverdb_read_le32( db ) + 0xA0C0;
#ifdef DEBUG
		fprintf( stderr, "%s: toc[%d] = 0x%08X (0x%08X)\n",
				 __func__, i + 1, offset - 0xA0C0, offset );
#endif /* DEBUG */
		toc = g_list_append( toc, ( void * ) offset );
	}

	/* Read in track info
	 */
	fseek( db, 0xA0C0, SEEK_SET );

	for( i = 0; i < total; i++ ) {
		
		assert( ftell( db ) == ( int ) g_list_nth_data( toc, i ) );

		data = ( iriverdb_data_t * ) calloc( sizeof( iriverdb_data_t ), 1 );

		/* Read field lengths
		 */
		for( j = 0; j < IRIVERDB_FIELD_MAX; j++ )
			field_length[j] = iriverdb_read_le16( db );

		/* Read fields
		 */
		for( j = 0; j < IRIVERDB_FIELD_MAX; j++ )
			data->field[j] = iriverdb_read_string( db,
							       field_length[j],
							       database->version );
#ifdef DEBUG
		fprintf( stderr, "%s: track %d = \"%s\"\n",
				 __func__, i + 1,
				 data->field[IRIVERDB_FIELD_FILENAME] );
		for( j = 1; j < IRIVERDB_FIELD_MAX; j++ )
			fprintf( stderr, "%s: field[%s] = \"%s\"\n",
					 __func__, field[j],
					 data->field[j] );
#endif /* DEBUG */

		database->tracks = g_list_append( database->tracks, data );
	}

	g_list_free( toc );

	fclose( db );

	return database;
}

/* Refresh the database found in the given filesystem root with no callback
 * for progress feedback
 */
int iriverdb_refresh( const char *path )
{
	return iriverdb_refresh_with_callback( path, NULL );
}

/* Refresh the database found in the given filesystem root using the given
 * callback to supply progress feedback
 */
int iriverdb_refresh_with_callback( const char *path,
				    void ( *callback )( const iriverdb_callback_t type,
							const void *data ) )
{
	char filename[PATH_MAX], *pathsep;
	FILE *db;
	struct stat buf;
	time_t mtime;
	iriverdb_t *database;
	int i, j, total;
	iriverdb_data_t *data;

	if ( ( database = iriverdb_read_with_callback( path, callback ) ) == NULL )
		return -1;

	sprintf( filename, "%s%c%s", path, G_DIR_SEPARATOR, DB_FILENAME );

	if ( stat( filename, &buf ) < 0 ) {
		return -1;
	}

	mtime = buf.st_mtime;

#ifdef DEBUG
	printf( "iriverdb_refresh_with_callback: modification time = %d\n", mtime );
#endif /* DEBUG */

	/* Delete any files from the database that don't exist anymore, and
	 * also delete any files that have a modification time later than that
	 * of the database so that they will be rescanned and added again
	 */
	for( i = 0; i < g_list_length( database->tracks ); i++ ) {

		data = g_list_nth_data( database->tracks, i );

		sprintf( filename, "%s%s", path, data->field[IRIVERDB_FIELD_FILENAME] );

#ifndef G_OS_WIN32
		/* Convert DOS-ish path back to Unix again
		 */
		while ( ( pathsep = strchr( filename, '\\' ) ) != NULL ) {
			*pathsep = G_DIR_SEPARATOR;
		}
#endif /* ! G_OS_WIN32 */

		if ( stat( filename, &buf ) < 0 ) {
			if ( errno == ENOENT ) {

				/* File doesn't exist, scrub it
				 */
#ifdef DEBUG
				fprintf( stderr, "%s: \"%s\" no longer exists\n",
						 __func__, data->field[IRIVERDB_FIELD_FILENAME] );
#endif /* DEBUG */
				for( j = 0; j < IRIVERDB_FIELD_MAX; j++ )
					if ( data->field[j] )
						free( data->field[j] );
				database->tracks = g_list_remove( database->tracks, data );
				free( data );
				i--;

				/* Call the callback
				 */
				if ( callback )
					callback( IRIVERDB_CALLBACK_EXPUNGE, ( const void * ) filename );
			} else {
				return -1;
			}
		} else if ( buf.st_mtime > mtime ) {

			/* File has been modified, scrub it so it will be
			 * added back in again
			 */
#ifdef DEBUG
			fprintf( stderr, "%s: \"%s\" has been modified (%d > %d)\n",
					 __func__, data->field[IRIVERDB_FIELD_FILENAME],
					 buf.st_mtime, mtime );
#endif /* DEBUG */
			for( j = 0; j < IRIVERDB_FIELD_MAX; j++ )
				if ( data->field[j] )
					free( data->field[j] );
			database->tracks = g_list_remove( database->tracks, data );
			free( data );
			i--;
		}
	}

	/* Process any files with a modification time later than that of the
	 * database itself
	 */
	total = g_list_length( database->tracks );
	total += iriverdb_recurse( &database->tracks, path, path, callback );
	assert( total == g_list_length( database->tracks ) );

	/* Write out the database again
	 */
	iriverdb_write_database( path, database, callback );

	iriverdb_free_database( database );

	/* XXX Return number of new files or grand total?
	 */
	return total;
}
