/* libiriverdb - file metadata extraction routines
 * Copyright 2005 Matt Dainty
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * $Id: formats.c 43 2006-12-27 10:36:27Z matt $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <glib.h>

#ifdef HAVE_VORBISFILE
# include <vorbis/vorbisfile.h>
#endif /* HAVE_VORBISFILE */

#ifdef HAVE_ID3TAG
# include <id3tag.h>
#endif /* HAVE_ID3TAG */

#include "internal.h"

iriverdb_data_t *iriverdb_ogg_info( char *filename );
iriverdb_data_t *iriverdb_mp3_info( char *filename );

iriverdb_format_t formats[] = {
#ifdef HAVE_VORBISFILE
	{ ".ogg", &iriverdb_ogg_info },
#endif /* HAVE_VORBISFILE */
#if defined HAVE_ID3TAG
	{ ".mp3", &iriverdb_mp3_info },
#endif /* HAVE_ID3TAG */
	{ NULL, NULL } /* End of list */
};

#ifdef HAVE_VORBISFILE

/* Scan Ogg Vorbis file, return pointer to data structure hopefully containing
 * all fields _except_ filename filled in. Null on failure
 */
iriverdb_data_t *iriverdb_ogg_info( char *filename )
{
	typedef struct iriverdb_vorbiscomments {
		int field;
		char *string;
		int length;
	} iriverdb_vorbiscomments_t;

	iriverdb_data_t *data;
	FILE *f;
	OggVorbis_File vf;
	vorbis_comment *vc;
	int i;
	const iriverdb_vorbiscomments_t *ptr;

	/* Map database fields to vorbis comments
	 */
	const iriverdb_vorbiscomments_t comments[] = {
		{ IRIVERDB_FIELD_SONGNAME, "title", 5 },
		{ IRIVERDB_FIELD_ARTIST, "artist", 6 },
		{ IRIVERDB_FIELD_ALBUM, "album", 5 },
		{ IRIVERDB_FIELD_GENRE, "genre", 5 },
		{ 0, NULL, 0 },
	};

#ifdef DEBUG
	fprintf( stderr, "%s: filename = \"%s\"\n",
			 __func__, filename );
#endif /* DEBUG */

	/* Open the file
	 */
	if ( ( f = fopen( filename, "r" ) ) == NULL ) {
		return NULL;
	}

	if ( ov_open( f, &vf, NULL, 0 ) < 0 ) {
		fclose( f );
		return NULL;
	}

	/* Retrieve the comments
	 */
	if ( ( vc = ov_comment( &vf, -1 ) ) == NULL ) {
		ov_clear( &vf );
		return NULL;
	}

	data = ( iriverdb_data_t * ) g_malloc0( sizeof( iriverdb_data_t ) );

	/* Iterate over the comments and copy any useful ones that we find
	 */
	for( i = 0; i < vc->comments; i++ ) {

#ifdef DEBUG
		fprintf( stderr, "%s: vorbis comment = \"%s\"\n",
				 __func__,  vc->user_comments[i] );
#endif /* DEBUG */

		for( ptr = comments; ptr->string != NULL; ptr++ ) {

			/* Vorbis comments _can_ appear multiple times, so
			 * for the time being, just use the first one we
			 * find
			 */
			if ( ( g_ascii_strncasecmp( vc->user_comments[i],
						    ptr->string,
						    ptr->length ) != 0 ) ||
			     ( data->field[ptr->field] != NULL ) ) {
				continue;
			}

			/* Copy everything after the comment field and '='.
			 * Vorbis comments are in UTF-8 so no converting
			 * required here
			 */
			data->field[ptr->field] = g_strdup( vc->user_comments[i] + ptr->length + 1 );
		}
	}

	/* Tidy up afterwards
	 */
	ov_clear( &vf );

	return data;
}

#endif /* HAVE_VORBISFILE */

#ifdef HAVE_ID3TAG

char *iriverdb_mp3_get_frame( const struct id3_tag *tag, const char *id )
{
	struct id3_frame *frame;
	union id3_field *field;
	const id3_ucs4_t *temp;
	char *string;

	if ( ( frame = id3_tag_findframe( tag, id, 0 ) ) == NULL ) {
		return NULL;
	}

	if ( ( field = id3_frame_field( frame, 1 ) ) == NULL ) {
		return NULL;
	}

	if ( id3_field_type( field ) != ID3_FIELD_TYPE_STRINGLIST ) {
		return NULL;
	}

	if ( ( temp = id3_field_getstrings( field, 0 ) ) == NULL ) {
		return NULL;
	}

#if G_BYTE_ORDER == G_LITTLE_ENDIAN
	string = iriverdb_ucs4le_to_utf8( ( const char * ) temp, NULL );
#elif G_BYTE_ORDER == G_BIG_ENDIAN /* ! G_LITTLE_ENDIAN */
	string = iriverdb_ucs4be_to_utf8( ( const char * ) temp, NULL );
#endif /* G_BIG_ENDIAN */

#ifdef DEBUG
	fprintf( stderr, "%s: id3 frame field \"%s\" = \"%s\"\n",
			 __func__, id, string );
#endif /* DEBUG */

	return string;
}

iriverdb_data_t *iriverdb_mp3_info( char *filename )
{
	iriverdb_data_t *data;
	struct id3_file *file;
	struct id3_tag *tag;
	char *temp, *ptr;
	int i, len;
	id3_ucs4_t *genre;

#ifdef DEBUG
	fprintf( stderr, "%s: filename = \"%s\"\n",
			 __func__, filename );
#endif /* DEBUG */

	if ( ( file = id3_file_open( filename, ID3_FILE_MODE_READONLY ) ) == NULL ) {
		return NULL;
	}

	data = ( iriverdb_data_t * ) g_malloc0( sizeof( iriverdb_data_t ) );

	tag = id3_file_tag( file );

	/* Standard text fields
	 */
	data->field[IRIVERDB_FIELD_SONGNAME] = iriverdb_mp3_get_frame( tag, ID3_FRAME_TITLE );
	data->field[IRIVERDB_FIELD_ARTIST] = iriverdb_mp3_get_frame( tag, ID3_FRAME_ARTIST );
	data->field[IRIVERDB_FIELD_ALBUM] = iriverdb_mp3_get_frame( tag, ID3_FRAME_ALBUM );

	if ( ( temp = iriverdb_mp3_get_frame( tag, ID3_FRAME_GENRE ) ) != NULL ) {

#ifdef DEBUG
		fprintf( stderr, "%s: raw genre = \"%s\"\n",
				 __func__, temp );
#endif /* DEBUG */

		if ( ( ( temp[0] == '(' ) &&
		       ( temp[strlen( temp ) - 1] == ')' ) &&
		       ( strspn( temp + 1, "0123456789" ) == ( strlen( temp ) - 2 ) ) ) ||
		     ( strspn( temp, "0123456789" ) == strlen( temp ) ) ) {

			/* If the TCON tag is either just "<number>" or
			 * "(<number>)" try and convert it into a proper
			 * genre
			 */
			ptr = ( temp[0] == '(' ) ? temp + 1 : temp;
			i = atoi( ptr );

#ifdef DEBUG
			fprintf( stderr, "%s: genre index = \"%s\"\n",
					 __func__, ptr );
#endif /* DEBUG */

			if ( ( genre = id3_genre_index( i ) ) != NULL ) {
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
				data->field[IRIVERDB_FIELD_GENRE] = iriverdb_ucs4le_to_utf8( ( const char * ) genre, NULL );
#elif G_BYTE_ORDER == G_BIG_ENDIAN /* ! G_LITTLE_ENDIAN */
				data->field[IRIVERDB_FIELD_GENRE] = iriverdb_ucs4be_to_utf8( ( const char * ) genre, NULL );
#endif
			}

			g_free( temp );
		} else {

			/* Otherwise just return the TCON tag "as-is"
			 */
			data->field[IRIVERDB_FIELD_GENRE] = temp;
		}


#ifdef DEBUG
		fprintf( stderr, "%s: genre = \"%s\"\n",
				 __func__,
				 data->field[IRIVERDB_FIELD_GENRE] );
#endif /* DEBUG */
	}

	id3_file_close( file );

	return data;
}

#endif /* HAVE_ID3TAG */
