/* libiriverdb - example utility
 * Copyright 2005 Matt Dainty
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * $Id: main.c 30 2005-08-14 16:11:02Z  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <locale.h>
#define _GNU_SOURCE
#include <getopt.h>
#include <glib.h>
#include <glib/gi18n.h>

#include "iriverdb.h"

#define PROGRAM_NAME "iriverdb"

#define AUTHOR "Matt Dainty"

char *program_name;

static int database = -1;

static struct option const long_options[] = {
	{ "help", no_argument, NULL, 'h' },
	{ "verbose", no_argument, NULL, 'v' },
	{ "version", no_argument, NULL, 'V' },
	{ "database", required_argument, NULL, 'd' },
	/* model/database version aliases
	 */
	{ "h100", no_argument, &database, 0 },
	{ "h300", no_argument, &database, 1 },
	{ NULL, 0, NULL, 0 }
};

void callback( const iriverdb_callback_t type, const void *data )
{
	switch( type ) {
		case IRIVERDB_CALLBACK_SCAN:
			fprintf( stdout, _( "\
Scanning \"%s\"\n\
" ), ( char * ) data );
			break;
		case IRIVERDB_CALLBACK_EXPUNGE:
			fprintf( stdout, _( "\
\"%s\" no longer exists\n\
" ), ( char * ) data );
			break;
		default:
			break;
	}
}

void usage( int status )
{
	if ( status != EXIT_SUCCESS ) {
		fprintf( stderr, _( "Try `%s --help' for more information.\n" ),
				 program_name );
	} else {
		fprintf( stdout, _( "\
Usage: %s [OPTION]... ACTION DIRECTORY\n\
" ), program_name );
		fputs( _( "\
Perform ACTION on iRiver jukebox mounted at DIRECTORY.\n\
\n\
  -d, --database=VERSION create specified version of the iRiver database.\n\
                         Either this option must be specified, or use one of\n\
                         the convenience options below for the specific\n\
                         jukebox model\n\
      --h300             iRiver 20/40GB H300 series\n\
      --h100             iRiver 10/20/40GB H100 series\n\
" ), stdout );
		fputs( _( "\
  -v, --verbose          be verbose\n\
" ), stdout );
		fputs( _( "\
  -h, --help             display this help and exit\n\
" ), stdout );
		fputs( _( "\
  -V, --version          output version information and exit\n\
" ), stdout );
		fputs( _( "\
\n\
ACTION can be one of `write', `refresh' or `read' to write a new database\n\
from scratch, refresh an existing database, or dump the contents of an\n\
existing database, respectively.\n\
\n\
VERSION can be one of `0' or `1' corresponding to the 0.12 `8-bit string'\n\
database found on H100 series jukeboxes, or the 0.16 `16-bit UTF string'\n\
database found on H300 series jukeboxes.\n\
" ), stdout );
		fprintf( stdout, _( "\
\n\
Report bugs to <%s>.\n\
" ), PACKAGE_BUGREPORT );
	}

	exit( status );
}

int main( int argc, char *argv[] )
{
	int c, verbose = 0, i;
	struct stat buf;
	char *action, *directory;
	iriverdb_t *db;
	iriverdb_data_t *data;

	program_name = argv[0];
	setlocale( LC_ALL, "" );
	bindtextdomain( GETTEXT_PACKAGE, LOCALEDIR );
	bind_textdomain_codeset( GETTEXT_PACKAGE, "UTF-8" );
	textdomain( GETTEXT_PACKAGE );

	while ( ( c = getopt_long( argc, argv, "hvVd:", long_options, NULL ) ) != -1 ) {
		switch ( c ) {
			case 0:
				break;
			case 'v':
				verbose = 1;
				break;
			case 'd':
				database = atoi( optarg );
				switch ( database ) {
					case IRIVERDB_012:
					case IRIVERDB_016:
						break;
					default:
						fprintf( stderr, _( "\
Unknown database format: %s\n\
" ), optarg );
						exit( EXIT_FAILURE );
				}
				break;
			case 'V':
				fprintf( stdout, _( "\
%s (%s) %s\n\
Written by %s.\n\
" ), PROGRAM_NAME, PACKAGE_NAME, VERSION, AUTHOR );
				exit( EXIT_SUCCESS );
				break;
			case 'h':
				usage( EXIT_SUCCESS );
				break;
			default:
				usage( EXIT_FAILURE );
		}
	}

	if ( ( argc - optind ) != 2 ) {
		fprintf( stderr, _( "\
%s: too few arguments\n\
" ), program_name );
		usage( EXIT_FAILURE );
	}

	if ( database == -1 ) {
		fprintf( stderr, _( "\
%s: database version not specified\n\
" ), program_name );
		usage( EXIT_FAILURE );
	}

	action = argv[optind++];
	directory = argv[optind++];

	if ( ( ( stat( directory, &buf ) ) < 0 ) ||
	     ( ! S_ISDIR( buf.st_mode ) ) ) {
		fprintf( stderr, _( "\
%s: `%s' not a directory\n\
" ), program_name, directory );
		usage( EXIT_FAILURE );
	}

	if ( strcmp( action, "write" ) == 0 ) {
		if ( verbose == 0 )
			iriverdb_write( directory, database );
		else
			iriverdb_write_with_callback( directory, database, &callback );
	} else if ( strcmp( action, "refresh" ) == 0 ) {
		if ( verbose == 0 )
			iriverdb_refresh( directory );
		else
			iriverdb_refresh_with_callback( directory, &callback );
	} else if ( strcmp( action, "read" ) == 0 ) {
		if ( verbose == 0 )
			db = iriverdb_read( directory );
		else
			db = iriverdb_read_with_callback( directory, &callback );
		if ( db != NULL ) {
			fprintf( stdout, _( "\
Database version: %s\n\
Tracks: %d\n\
" ), ( db->version == IRIVERDB_012 ) ? "0.12" : "0.16",
   g_list_length( db->tracks ) );
			for( i = 0; i < g_list_length( db->tracks ); i++ ) {
				data = g_list_nth_data( db->tracks, i );
				fprintf( stdout, "\
%5d = %s\n\
", i + 1, data->field[IRIVERDB_FIELD_FILENAME] );
			}
			iriverdb_free_database( db );
		}
	} else {
		fprintf( stderr, _( "\
%s: unknown action `%s'\n\
" ), program_name, action );
		usage( EXIT_FAILURE );
	}

	exit( EXIT_SUCCESS );
}
